package sun.rmi.transport;

import java.lang.ref.PhantomReference;
import java.security.PrivilegedAction;
import sun.rmi.runtime.Log;
import java.io.InvalidClassException;
import java.rmi.UnmarshalException;
import java.rmi.dgc.Lease;
import java.util.Iterator;
import sun.rmi.runtime.NewThreadAction;
import java.rmi.RemoteException;
import sun.rmi.server.Util;
import sun.rmi.server.UnicastRef;
import java.util.HashSet;
import java.util.HashMap;
import sun.misc.GC;
import java.lang.ref.ReferenceQueue;
import java.util.Set;
import java.util.Map;
import java.rmi.dgc.DGC;
import java.security.ProtectionDomain;
import java.net.SocketPermission;
import java.security.Permissions;
import java.security.AccessController;
import sun.security.action.GetLongAction;
import java.util.List;
import java.security.AccessControlContext;
import java.rmi.server.ObjID;
import java.rmi.dgc.VMID;

final class DGCClient
{
    private static long nextSequenceNum;
    private static VMID vmid;
    private static final long leaseValue;
    private static final long cleanInterval;
    private static final long gcInterval;
    private static final int dirtyFailureRetries = 5;
    private static final int cleanFailureRetries = 5;
    private static final ObjID[] emptyObjIDArray;
    private static final ObjID dgcID;
    private static final AccessControlContext SOCKET_ACC;
    
    private DGCClient() {
    }
    
    static void registerRefs(Endpoint endpoint, List<LiveRef> list) {
        while (!DGCClient.EndpointEntry.lookup(endpoint).registerRefs(list)) {}
    }
    
    private static synchronized long getNextSequenceNum() {
        return DGCClient.nextSequenceNum++;
    }
    
    private static long computeRenewTime(long n, long n2) {
        return n + n2 / 2L;
    }
    
    static long access$300() {
        return getNextSequenceNum();
    }
    
    static long access$700(long n, long n2) {
        return computeRenewTime(n, n2);
    }
    
    static {
        DGCClient.nextSequenceNum = Long.MIN_VALUE;
        DGCClient.vmid = new VMID();
        leaseValue = (long)AccessController.doPrivileged(new GetLongAction("java.rmi.dgc.leaseValue", 600000L));
        cleanInterval = (long)AccessController.doPrivileged(new GetLongAction("sun.rmi.dgc.cleanInterval", 180000L));
        gcInterval = (long)AccessController.doPrivileged(new GetLongAction("sun.rmi.dgc.client.gcInterval", 3600000L));
        emptyObjIDArray = new ObjID[0];
        dgcID = new ObjID(2);
        Permissions permissions = new Permissions();
        permissions.add(new SocketPermission("*", "connect,resolve"));
        SOCKET_ACC = new AccessControlContext(new ProtectionDomain[] { new ProtectionDomain(null, permissions) });
    }
    
    private static class EndpointEntry
    {
        private Endpoint endpoint;
        private DGC dgc;
        private Map<LiveRef, DGCClient.EndpointEntry.RefEntry> refTable;
        private Set<DGCClient.EndpointEntry.RefEntry> invalidRefs;
        private boolean removed;
        private long renewTime;
        private long expirationTime;
        private int dirtyFailures;
        private long dirtyFailureStartTime;
        private long dirtyFailureDuration;
        private Thread renewCleanThread;
        private boolean interruptible;
        private ReferenceQueue<LiveRef> refQueue;
        private Set<DGCClient.EndpointEntry.CleanRequest> pendingCleans;
        private static Map<Endpoint, DGCClient.EndpointEntry> endpointTable;
        private static GC.LatencyRequest gcLatencyRequest;
        
        public static DGCClient.EndpointEntry lookup(Endpoint endpoint) {
            synchronized (DGCClient.EndpointEntry.endpointTable) {
                DGCClient.EndpointEntry endpointEntry = (DGCClient.EndpointEntry)DGCClient.EndpointEntry.endpointTable.get(endpoint);
                if (endpointEntry == null) {
                    endpointEntry = new EndpointEntry(endpoint);
                    DGCClient.EndpointEntry.endpointTable.put(endpoint, endpointEntry);
                    if (DGCClient.EndpointEntry.gcLatencyRequest == null) {
                        DGCClient.EndpointEntry.gcLatencyRequest = GC.requestLatency(DGCClient.gcInterval);
                    }
                }
                return endpointEntry;
            }
        }
        
        private EndpointEntry(Endpoint endpoint) {
            this.refTable = new HashMap(5);
            this.invalidRefs = new HashSet(5);
            this.removed = false;
            this.renewTime = Long.MAX_VALUE;
            this.expirationTime = Long.MIN_VALUE;
            this.dirtyFailures = 0;
            this.interruptible = false;
            this.refQueue = new ReferenceQueue();
            this.pendingCleans = new HashSet(5);
            this.endpoint = endpoint;
            try {
                this.dgc = (DGC)Util.createProxy(DGCImpl.class, new UnicastRef(new LiveRef(DGCClient.dgcID, endpoint, false)), true);
            }
            catch (RemoteException ex) {
                throw new Error("internal error creating DGC stub");
            }
            (this.renewCleanThread = (Thread)AccessController.doPrivileged(new NewThreadAction(new DGCClient.EndpointEntry.RenewCleanThread(), "RenewClean-" + endpoint, true))).start();
        }
        
        public boolean registerRefs(List<LiveRef> list) {
            assert !Thread.holdsLock(this);
            Object o = null;
            long access$300;
            synchronized (this) {
                if (this.removed) {
                    return false;
                }
                for (LiveRef liveRef : list) {
                    assert !(!liveRef.getEndpoint().equals(this.endpoint));
                    DGCClient.EndpointEntry.RefEntry refEntry = (DGCClient.EndpointEntry.RefEntry)this.refTable.get(liveRef);
                    if (refEntry == null) {
                        LiveRef liveRef2 = (LiveRef)liveRef.clone();
                        refEntry = new DGCClient.EndpointEntry.RefEntry(liveRef2);
                        this.refTable.put(liveRef2, refEntry);
                        if (o == null) {
                            o = new HashSet(5);
                        }
                        o.add(refEntry);
                    }
                    refEntry.addInstanceToRefSet(liveRef);
                }
                if (o == null) {
                    return true;
                }
                o.addAll(this.invalidRefs);
                this.invalidRefs.clear();
                access$300 = DGCClient.access$300();
            }
            this.makeDirtyCall(o, access$300);
            return true;
        }
        
        private void removeRefEntry(DGCClient.EndpointEntry.RefEntry refEntry) {
            assert !(!Thread.holdsLock(this));
            assert !this.removed;
            assert !(!this.refTable.containsKey(refEntry.getRef()));
            this.refTable.remove(refEntry.getRef());
            this.invalidRefs.remove(refEntry);
            if (this.refTable.isEmpty()) {
                synchronized (DGCClient.EndpointEntry.endpointTable) {
                    DGCClient.EndpointEntry.endpointTable.remove(this.endpoint);
                    this.endpoint.getOutboundTransport().free(this.endpoint);
                    if (DGCClient.EndpointEntry.endpointTable.isEmpty()) {
                        assert DGCClient.EndpointEntry.gcLatencyRequest != null;
                        DGCClient.EndpointEntry.gcLatencyRequest.cancel();
                        DGCClient.EndpointEntry.gcLatencyRequest = null;
                    }
                    this.removed = true;
                }
            }
            return;
        }
        
        private void makeDirtyCall(Set<DGCClient.EndpointEntry.RefEntry> set, long n) {
            assert !Thread.holdsLock(this);
            ObjID[] array;
            if (set != null) {
                array = createObjIDArray(set);
            }
            else {
                array = DGCClient.emptyObjIDArray;
            }
            long currentTimeMillis = System.currentTimeMillis();
            try {
                long value = this.dgc.dirty(array, n, new Lease(DGCClient.vmid, DGCClient.leaseValue)).getValue();
                long access$700 = DGCClient.access$700(currentTimeMillis, value);
                long expirationTime = currentTimeMillis + value;
                synchronized (this) {
                    this.dirtyFailures = 0;
                    this.setRenewTime(access$700);
                    this.expirationTime = expirationTime;
                }
            }
            catch (Exception ex) {
                long currentTimeMillis2 = System.currentTimeMillis();
                synchronized (this) {
                    ++this.dirtyFailures;
                    if (ex instanceof UnmarshalException && ex.getCause() instanceof InvalidClassException) {
                        DGCImpl.dgcLog.log(Log.BRIEF, "InvalidClassException exception in DGC dirty call", ex);
                        return;
                    }
                    if (this.dirtyFailures == 1) {
                        this.dirtyFailureStartTime = currentTimeMillis;
                        this.dirtyFailureDuration = currentTimeMillis2 - currentTimeMillis;
                        this.setRenewTime(currentTimeMillis2);
                    }
                    else {
                        int n2 = this.dirtyFailures - 2;
                        if (n2 == 0) {
                            this.dirtyFailureDuration = Math.max(this.dirtyFailureDuration + (currentTimeMillis2 - currentTimeMillis) >> 1, 1000L);
                        }
                        long renewTime = currentTimeMillis2 + (this.dirtyFailureDuration << n2);
                        if (renewTime < this.expirationTime || this.dirtyFailures < 5 || renewTime < this.dirtyFailureStartTime + DGCClient.leaseValue) {
                            this.setRenewTime(renewTime);
                        }
                        else {
                            this.setRenewTime(Long.MAX_VALUE);
                        }
                    }
                    if (set != null) {
                        this.invalidRefs.addAll(set);
                        Iterator<Object> iterator = set.iterator();
                        while (iterator.hasNext()) {
                            ((DGCClient.EndpointEntry.RefEntry)iterator.next()).markDirtyFailed();
                        }
                    }
                    if (this.renewTime >= this.expirationTime) {
                        this.invalidRefs.addAll(this.refTable.values());
                    }
                    return;
                }
            }
            return;
        }
        
        private void setRenewTime(long n) {
            assert !(!Thread.holdsLock(this));
            if (n < this.renewTime) {
                this.renewTime = n;
                if (this.interruptible) {
                    AccessController.doPrivileged(new PrivilegedAction<Void>() {
                        @Override
                        public Void run() {
                            DGCClient.EndpointEntry.this.renewCleanThread.interrupt();
                            return null;
                        }
                    });
                }
            }
            else {
                this.renewTime = n;
            }
            return;
        }
        
        private void processPhantomRefs(DGCClient.EndpointEntry.RefEntry.PhantomLiveRef phantomLiveRef) {
            assert !(!Thread.holdsLock(this));
            Object o = null;
            Object o2 = null;
            do {
                DGCClient.EndpointEntry.RefEntry refEntry = phantomLiveRef.getRefEntry();
                refEntry.removeInstanceFromRefSet(phantomLiveRef);
                if (refEntry.isRefSetEmpty()) {
                    if (refEntry.hasDirtyFailed()) {
                        if (o == null) {
                            o = new HashSet(5);
                        }
                        o.add(refEntry);
                    }
                    else {
                        if (o2 == null) {
                            o2 = new HashSet(5);
                        }
                        o2.add(refEntry);
                    }
                    this.removeRefEntry(refEntry);
                }
                else {
                    continue;
                }
            } while ((phantomLiveRef = (DGCClient.EndpointEntry.RefEntry.PhantomLiveRef)this.refQueue.poll()) != null);
            if (o != null) {
                this.pendingCleans.add(new DGCClient.EndpointEntry.CleanRequest(createObjIDArray(o), DGCClient.access$300(), true));
            }
            if (o2 != null) {
                this.pendingCleans.add(new DGCClient.EndpointEntry.CleanRequest(createObjIDArray(o2), DGCClient.access$300(), false));
            }
            return;
        }
        
        private void makeCleanCalls() {
            assert !Thread.holdsLock(this);
            Iterator<DGCClient.EndpointEntry.CleanRequest> iterator = this.pendingCleans.iterator();
            while (iterator.hasNext()) {
                DGCClient.EndpointEntry.CleanRequest cleanRequest = (DGCClient.EndpointEntry.CleanRequest)iterator.next();
                try {
                    this.dgc.clean(cleanRequest.objIDs, cleanRequest.sequenceNum, DGCClient.vmid, cleanRequest.strong);
                    iterator.remove();
                }
                catch (Exception ex) {
                    int failures = cleanRequest.failures + 1;
                    cleanRequest.failures = failures;
                    if (failures >= 5) {
                        iterator.remove();
                    }
                    else {
                        continue;
                    }
                }
            }
            return;
        }
        
        private static ObjID[] createObjIDArray(Set<DGCClient.EndpointEntry.RefEntry> set) {
            ObjID[] array = new ObjID[set.size()];
            Iterator<Object> iterator = set.iterator();
            for (int i = 0; i < array.length; ++i) {
                array[i] = ((DGCClient.EndpointEntry.RefEntry)iterator.next()).getRef().getObjID();
            }
            return array;
        }
        
        static void access$1400(DGCClient.EndpointEntry endpointEntry, DGCClient.EndpointEntry.RefEntry.PhantomLiveRef phantomLiveRef) {
            endpointEntry.processPhantomRefs(phantomLiveRef);
        }
        
        static void access$1600(DGCClient.EndpointEntry endpointEntry, Set set, long n) {
            endpointEntry.makeDirtyCall(set, n);
        }
        
        static void access$1700(DGCClient.EndpointEntry endpointEntry) {
            endpointEntry.makeCleanCalls();
        }
        
        static {
            DGCClient.EndpointEntry.endpointTable = new HashMap(5);
            DGCClient.EndpointEntry.gcLatencyRequest = null;
        }
        
        private class RenewCleanThread implements Runnable
        {
            @Override
            public void run() {
                do {
                    DGCClient.EndpointEntry.RefEntry.PhantomLiveRef phantomLiveRef = null;
                    boolean n = false;
                    Set<Object> access$1500 = null;
                    long access$1501 = Long.MIN_VALUE;
                    long n2;
                    synchronized (DGCClient.EndpointEntry.this) {
                        n2 = Math.max(DGCClient.EndpointEntry.this.renewTime - System.currentTimeMillis(), 1L);
                        if (!DGCClient.EndpointEntry.this.pendingCleans.isEmpty()) {
                            n2 = Math.min(n2, DGCClient.cleanInterval);
                        }
                        DGCClient.EndpointEntry.this.interruptible = true;
                    }
                    try {
                        phantomLiveRef = (DGCClient.EndpointEntry.RefEntry.PhantomLiveRef)DGCClient.EndpointEntry.this.refQueue.remove(n2);
                    }
                    catch (InterruptedException ex) {}
                    synchronized (DGCClient.EndpointEntry.this) {
                        DGCClient.EndpointEntry.this.interruptible = false;
                        Thread.interrupted();
                        if (phantomLiveRef != null) {
                            DGCClient.EndpointEntry.access$1400(DGCClient.EndpointEntry.this, phantomLiveRef);
                        }
                        if (System.currentTimeMillis() > DGCClient.EndpointEntry.this.renewTime) {
                            n = true;
                            if (!DGCClient.EndpointEntry.this.invalidRefs.isEmpty()) {
                                access$1500 = DGCClient.EndpointEntry.this.invalidRefs;
                                DGCClient.EndpointEntry.this.invalidRefs = new HashSet(5);
                            }
                            access$1501 = DGCClient.access$300();
                        }
                    }
                    AccessController.doPrivileged(new PrivilegedAction<Void>() {
                        @Override
                        public Void run() {
                            if (!(n == 0)) {
                                DGCClient.EndpointEntry.access$1600(DGCClient.EndpointEntry.this, access$1500, access$1501);
                            }
                            if (!DGCClient.EndpointEntry.this.pendingCleans.isEmpty()) {
                                DGCClient.EndpointEntry.access$1700(DGCClient.EndpointEntry.this);
                            }
                            return null;
                        }
                    }, DGCClient.SOCKET_ACC);
                } while (!DGCClient.EndpointEntry.this.removed || !DGCClient.EndpointEntry.this.pendingCleans.isEmpty());
            }
        }
        
        private static class CleanRequest
        {
            final ObjID[] objIDs;
            final long sequenceNum;
            final boolean strong;
            int failures;
            
            CleanRequest(ObjID[] objIDs, long sequenceNum, boolean strong) {
                this.failures = 0;
                this.objIDs = objIDs;
                this.sequenceNum = sequenceNum;
                this.strong = strong;
            }
        }
        
        private class RefEntry
        {
            private LiveRef ref;
            private Set<DGCClient.EndpointEntry.RefEntry.PhantomLiveRef> refSet;
            private boolean dirtyFailed;
            
            public RefEntry(LiveRef ref) {
                this.refSet = new HashSet(5);
                this.dirtyFailed = false;
                this.ref = ref;
            }
            
            public LiveRef getRef() {
                return this.ref;
            }
            
            public void addInstanceToRefSet(LiveRef liveRef) {
                assert !(!Thread.holdsLock(DGCClient.EndpointEntry.this));
                assert !(!liveRef.equals(this.ref));
                this.refSet.add(new DGCClient.EndpointEntry.RefEntry.PhantomLiveRef(liveRef));
                return;
            }
            
            public void removeInstanceFromRefSet(DGCClient.EndpointEntry.RefEntry.PhantomLiveRef phantomLiveRef) {
                assert !(!Thread.holdsLock(DGCClient.EndpointEntry.this));
                assert !(!this.refSet.contains(phantomLiveRef));
                this.refSet.remove(phantomLiveRef);
                return;
            }
            
            public boolean isRefSetEmpty() {
                assert !(!Thread.holdsLock(DGCClient.EndpointEntry.this));
                return this.refSet.size() == 0;
            }
            
            public void markDirtyFailed() {
                assert !(!Thread.holdsLock(DGCClient.EndpointEntry.this));
                this.dirtyFailed = true;
                return;
            }
            
            public boolean hasDirtyFailed() {
                assert !(!Thread.holdsLock(DGCClient.EndpointEntry.this));
                return this.dirtyFailed;
            }
            
            private class PhantomLiveRef extends PhantomReference<LiveRef>
            {
                public PhantomLiveRef(LiveRef referent) {
                    super(referent, DGCClient.EndpointEntry.RefEntry.this.this$0.refQueue);
                }
                
                public DGCClient.EndpointEntry.RefEntry getRefEntry() {
                    return DGCClient.EndpointEntry.RefEntry.this;
                }
            }
        }
    }
}
